package flink.zanxiangnet.ad.monitoring;

import flink.zanxiangnet.ad.monitoring.pojo.entity.*;
import flink.zanxiangnet.ad.monitoring.process.PlanHourDTOStreamProcess;
import flink.zanxiangnet.ad.monitoring.process.PlanHourDWDProcess;
import flink.zanxiangnet.ad.monitoring.process.PlanHourStreamCompletionProcess;
import flink.zanxiangnet.ad.monitoring.process.PlanMinuteDWDProcess;
import flink.zanxiangnet.ad.monitoring.sink.TunnelBatchStreamSink;
import flink.zanxiangnet.ad.monitoring.stream.KeyedBatchStream;
import flink.zanxiangnet.ad.monitoring.trigger.PlanMinuteODSStreamTrigger;
import flink.zanxiangnet.ad.monitoring.kafka.KafkaComponent;
import org.apache.commons.lang3.StringUtils;
import org.apache.flink.api.common.eventtime.SerializableTimestampAssigner;
import org.apache.flink.api.common.eventtime.WatermarkStrategy;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.connector.kafka.source.KafkaSource;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.DataStreamSource;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.windowing.assigners.TumblingEventTimeWindows;
import org.apache.flink.streaming.api.windowing.time.Time;
import org.apache.flink.util.OutputTag;

import java.time.Duration;
import java.util.*;

public class PlanHourStreamJob {

    /**
     * 可能有数据的最早日期
     */
    private static final String OLDEST_DAY = "2019-01-01";

    public static void main(String[] args) throws Exception {
        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();

        // 加载配置文件到 flink的全局配置中
        Properties props = new Properties();
        props.load(PlanHourStreamJob.class.getResourceAsStream("/application.properties"));
        Configuration configuration = new Configuration();
        props.stringPropertyNames().forEach(key -> {
            String value = props.getProperty(key);
            configuration.setString(key.trim(), StringUtils.isBlank(value) ? "" : value.trim());
        });
        env.getConfig().setGlobalJobParameters(configuration);

        KafkaSource<String> adStreamOfMinuteSource = KafkaComponent.buildKafkaSource(props, KafkaComponent.KafkaTopic.adHourTopic, KafkaComponent.KafkaTopic.KafkaGroupId.planHourConsumerGroup);
        DataStreamSource<String> adStreamOfMinuteIn = env.fromSource(adStreamOfMinuteSource, WatermarkStrategy.noWatermarks(), "planHourSource_kafka");

        // 广告分钟数据（前 5分钟的广告消耗数据）
        final OutputTag<AdDataOfMinuteODS> adMinuteStreamTag = new OutputTag<AdDataOfMinuteODS>("adMinuteStream") {
        };
        // 广告小时数据（往前回滚 10天）
        final OutputTag<AdDataOfHourODS> adHourStreamTag = new OutputTag<AdDataOfHourODS>("adHourStream") {
        };
        // 计划小时数据（从分钟数据流切出来的整点数据）
        final OutputTag<PlanStatOfHourDWD> planHourFromMinuteStreamTag = new OutputTag<PlanStatOfHourDWD>("planHourFromMinuteStream") {
        };

        // 对流进行映射，拆分（实时的分钟流和回滚的小时流）
        SingleOutputStreamOperator<AdDataOfMinuteODS> adODSStream = adStreamOfMinuteIn.filter(StringUtils::isNotBlank)
                .process(new PlanHourDTOStreamProcess(adMinuteStreamTag, adHourStreamTag));

        // 分钟流-计算
        SingleOutputStreamOperator<PlanStatOfMinuteDWD> planMinuteDWDStream = adODSStream.getSideOutput(adMinuteStreamTag)
                // 打水印，允许数据延迟 6分钟，同时指定时间流
                .assignTimestampsAndWatermarks(WatermarkStrategy.<AdDataOfMinuteODS>forBoundedOutOfOrderness(Duration.ofMinutes(6L))
                        .withTimestampAssigner((SerializableTimestampAssigner<AdDataOfMinuteODS>) (adODS, l) -> adODS.getStatTime()))
                .keyBy(AdDataOfMinuteODS::getCampaignId)
                // 开一个 5分钟的滚动窗口
                .window(TumblingEventTimeWindows.of(Time.minutes(5L)))
                .trigger(new PlanMinuteODSStreamTrigger())
                .process(new PlanMinuteDWDProcess(planHourFromMinuteStreamTag));
        new KeyedBatchStream<>("planMinuteDWDStream", planMinuteDWDStream.keyBy(PlanStatOfMinuteDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(PlanStatOfMinuteDWD.class))
                .name("sink_plan_minute_dwd");

        // 小时流
        SingleOutputStreamOperator<PlanStatOfHourDWD> planHourDWDStream = adODSStream.getSideOutput(adHourStreamTag).keyBy(AdDataOfHourODS::getCampaignId)
                .process(new PlanHourDWDProcess());
        DataStream<PlanStatOfHourDWD> planHourDWDAllStream = planMinuteDWDStream.getSideOutput(planHourFromMinuteStreamTag)
                .keyBy(PlanStatOfHourDWD::getCampaignId)
                .process(new PlanHourStreamCompletionProcess())
                .union(planHourDWDStream);
        new KeyedBatchStream<>("planHourDWDStream", planHourDWDAllStream.keyBy(PlanStatOfHourDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(PlanStatOfHourDWD.class))
                .name("sink_plan_hour_dwd");

        env.execute("plan_hour_stream_job");
    }
}
