package flink.zanxiangnet.ad.monitoring.process;

import flink.zanxiangnet.ad.monitoring.config.ClickhouseDataSourceFactory;
import flink.zanxiangnet.ad.monitoring.dao.mapper.AdStatOfDayDWDMapper;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdDataOfHourODS;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfDayDWD;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfHourDWD;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.util.DateUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.state.MapState;
import org.apache.flink.api.common.state.MapStateDescriptor;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.api.common.typeinfo.Types;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.KeyedProcessFunction;
import org.apache.flink.util.Collector;
import org.apache.ibatis.datasource.DataSourceFactory;
import org.apache.ibatis.mapping.Environment;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.ibatis.transaction.jdbc.JdbcTransactionFactory;

import java.time.LocalDate;
import java.util.Map;
import java.util.Properties;
import java.util.stream.Collectors;

/**
 * 往前回滚 10天的小时数据（警告，此数据依赖小时数据必定在天数据之后拉取来实现的。不然会拉下昨天的数据没有统计进来）
 */
@Slf4j
public class AdHourDWDProcess extends KeyedProcessFunction<Long, AdDataOfHourODS, AdStatOfHourDWD> {
    private SqlSessionFactory sqlSessionFactory;
    // 上次查询的天数据
    private ValueState<String> lastQueryDayState;
    // 聚合的天的数据
    private MapState<String, AdStatOfDayDWD> historyReduceState;
    // 上小时聚合的结果
    private MapState<String, AdStatOfHourDWD> lastReduceState;

    @Override
    public void open(Configuration conf) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();

        Properties ckProps = new Properties();
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_URL, params.get(ApplicationProperties.CK_URL));
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_USER, params.get(ApplicationProperties.CK_USERNAME));
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_PASSWORD, params.get(ApplicationProperties.CK_PASSWORD));

        DataSourceFactory dataSourceFactory = new ClickhouseDataSourceFactory();
        dataSourceFactory.setProperties(ckProps);
        Environment environment = new Environment("clickhouse", new JdbcTransactionFactory(), dataSourceFactory.getDataSource());
        org.apache.ibatis.session.Configuration configuration = new org.apache.ibatis.session.Configuration(environment);
        // 开启驼峰规则
        configuration.setMapUnderscoreToCamelCase(true);
        configuration.getTypeAliasRegistry().registerAlias(AdStatOfDayDWD.class);
        // addMapper一定要放到 alias的后面！！！！！
        configuration.addMapper(AdStatOfDayDWDMapper.class);
        sqlSessionFactory = new SqlSessionFactoryBuilder().build(configuration);

        lastQueryDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastQueryDayState", String.class));
        historyReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("historyReduceState", Types.STRING, Types.POJO(AdStatOfDayDWD.class)));
        lastReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("lastReduceState", Types.STRING, Types.POJO(AdStatOfHourDWD.class)));
    }

    @Override
    public void processElement(AdDataOfHourODS element, KeyedProcessFunction<Long, AdDataOfHourODS, AdStatOfHourDWD>.Context context, Collector<AdStatOfHourDWD> collector) throws Exception {
        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
        long now = System.currentTimeMillis();
        LocalDate today = LocalDate.now();
        long adId = element.getAdId();

        String lastQueryDay = lastQueryDayState.value();
        // 从 maxCompute拉取指定 广告的历史数据
        if (lastQueryDay == null || !lastQueryDay.equals(DateUtil.formatLocalDate(today))) {
            try (SqlSession session = sqlSessionFactory.openSession()) {
                AdStatOfDayDWDMapper mapper = session.getMapper(AdStatOfDayDWDMapper.class);
                Map<String, AdStatOfDayDWD> historyData = mapper.lastReduceResult(adId, null, DateUtil.formatLocalDate(LocalDate.now().minusDays(1L)), 60).stream()
                        .collect(Collectors.toMap(AdStatOfDayDWD::getStatDay, data -> data, (val1, val2) -> val2));
                historyReduceState.clear();
                historyReduceState.putAll(historyData);
                lastQueryDayState.update(DateUtil.formatLocalDate(today));
            }
        }
        AdStatOfDayDWD yesterdayReduceData = null;
        for (int i = 1; i < 60; i++) {
            LocalDate day = statDay.minusDays(i);
            yesterdayReduceData = historyReduceState.get(DateUtil.formatLocalDate(day));
            if (yesterdayReduceData != null) {
                break;
            }
        }

        AdStatOfHourDWD lastReduce = lastReduceState.get(element.getStatDay());

        AdStatOfHourDWD newStatData = AdStatOfHourDWD.reduce(yesterdayReduceData, lastReduce, element, now);
        lastReduceState.put(newStatData.getStatDay(), newStatData);
        collector.collect(newStatData);

        // 往前清理 15天的数据
        lastReduceState.remove(DateUtil.formatLocalDate(statDay.minusDays(15L)));
    }
}
